module ACTPol_CMBonly 

  implicit none
  private
  integer, parameter :: campc = KIND(1.d0)

  character(LEN=*), parameter, public :: data_dir='data/' !set path to data directory
  character(LEN=*), parameter, public :: ap_like='ACTPol_dr4_cmbonly_like'

  !Possible combinations: TT only, TE only, EE only, TT+TE+EE
  ! i.e., true-false-false, false-true-false, false-false-true, true-true-true
  logical :: use_tt  = .true.
  logical :: use_te  = .true.
  logical :: use_ee  = .true.
 
  integer, public :: tt_lmax = 6000
  integer, public :: bmin = 1 ! 1 for ACTPol only or ACTPol+WMAP, 24 for ACTPol+Planck

  integer, public, parameter :: b0 = 6 !first bin in TT theory selection, don't change
  integer, public, parameter :: nbin = 260 ! total bins   
  integer, public, parameter :: nbinw = 130 !total bins in single patch
  integer, public, parameter :: nbintt = 40
  integer, public, parameter :: nbinte = 45
  integer, public, parameter :: nbinee = 45
  integer, public, parameter :: lmax_win = 7925 !total ell in window functions
  integer, public, parameter :: bmax_win = 520 !total bins in window functions
  integer, public, parameter :: bmax = 52 !total bins in windows for one spec 
  !-------------------------------------------------------
  real(campc), parameter :: PI    = 3.14159265358979323846264d0
  real(campc), dimension(:), allocatable ::  bval,X_data,X_sig,diff_vec
  real(campc), dimension(:,:), allocatable :: win_func_d,win_func_w
  real(campc), dimension(:,:), allocatable :: covmat, fisher
 
  public actpol_like_init, actpol_calc_like, get_free_lun
  contains
  
  ! ===========================================================================
  subroutine actpol_like_init
 
  integer  :: i,j,lun,il,info,dum,k,bin_no
  character(LEN=1024) :: like_file, cov_file, bbld_file, bblw_file
  logical  :: good
    
  print *, 'Initializing ACTPol likelihood, version '//ap_like
 
  like_file    = trim(data_dir)//'cl_cmb_ap.dat'
  cov_file     = trim(data_dir)//'c_matrix_ap.dat'
  bbld_file    = trim(data_dir)//'coadd_bpwf_15mJy_191127_lmin2.txt'
  bblw_file    = trim(data_dir)//'coadd_bpwf_100mJy_191127_lmin2.txt'

  allocate(bval(nbin),X_data(nbin),X_sig(nbin))
  allocate(covmat(nbin,nbin))
   
  inquire(file=like_file, exist=good)
  if(.not.good)then
     write(*,*) 'file not found', trim(like_file), trim(data_dir)
     stop
  endif
   
  call get_free_lun(lun)
  open(unit=lun,file=like_file,form='formatted',status='unknown',action='read')
  do i=1,nbin !read TT,TE,EE
     read(lun,*) bval(i),X_data(i),X_sig(i)
  enddo
  close(lun)

  inquire(file=cov_file, exist=good)
  if(.not.good)then
     write(*,*) 'file not found', trim(cov_file), trim(data_dir)
     stop
  endif
  call get_free_lun(lun)
  open(unit=lun,file=cov_file,form='unformatted',status='old')
      read(lun) covmat
  close(lun)
  do i=1,nbin
     do j=i+1,nbin
        covmat(j,i)=covmat(i,j) !covmat is now full matrix
     enddo
  enddo

!select lmin in TT for combination with WMAP/Planck
  if (bmin /= 1) then
        do i=1,bmin
                covmat(i,1:nbintt)=0.d0 !deep
                covmat(1:nbintt,i)=0.d0 !deep
                covmat(i,i)=1d+10       !deep
                covmat(nbinw+i,nbinw+1:nbinw+nbintt)=0.d0 !wide
                covmat(nbinw+1:nbinw+nbintt,nbinw+i)=0.d0 !wide
                covmat(nbinw+i,nbinw+i)=1d+10             !wide
        end do
  end if

 !Select covmat
 !Only TT
  if((use_tt .eqv. .true.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .false.)) then
       bin_no=nbintt+nbintt
       allocate(fisher(bin_no,bin_no))
       fisher(1:nbintt,1:nbintt)=covmat(1:nbintt,1:nbintt) !deep
       fisher(nbintt+1:bin_no,nbintt+1:bin_no)=covmat(nbinw+1:nbinw+nbintt,nbinw+1:nbinw+nbintt) !wide
       fisher(1:nbintt,nbintt+1:bin_no)=covmat(1:nbintt,nbinw+1:nbinw+nbintt) !offdiag
       fisher(nbintt+1:bin_no,1:nbintt)=covmat(nbinw+1:nbinw+nbintt,1:nbintt) !offdiag
  !Only TE
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .false.)) then
       bin_no=nbinte+nbinte
       allocate(fisher(bin_no,bin_no))
       fisher(1:nbinte,1:nbinte)=covmat(nbintt+1:nbintt+nbinte,nbintt+1:nbintt+nbinte) !deep
       fisher(nbinte+1:bin_no,nbinte+1:bin_no)=covmat(nbinw+nbintt+1:nbinw+nbintt+nbinte,nbinw+nbintt+1:nbinw+nbintt+nbinte) !wide
       fisher(1:nbinte,nbinte+1:bin_no)=covmat(nbintt+1:nbintt+nbinte,nbinw+nbintt+1:nbinw+nbintt+nbinte) !offdiag
       fisher(nbinte+1:bin_no,1:nbinte)=covmat(nbinw+nbintt+1:nbinw+nbintt+nbinte,nbintt+1:nbintt+nbinte) !offdiag
  !Only EE 
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbinee+nbinee
       allocate(fisher(bin_no,bin_no))
       fisher(1:nbinee,1:nbinee)=covmat(nbintt+nbinte+1:nbintt+nbinte+nbinee,nbintt+nbinte+1:nbintt+nbinte+nbinee) !deep
       fisher(nbinee+1:bin_no,nbinee+1:bin_no)=covmat(nbinw+nbintt+nbinte+1:nbinw+nbinw,nbinw+nbintt+nbinte+1:nbinw+nbinw) !wide
       fisher(1:nbinee,nbinee+1:bin_no)=covmat(nbintt+nbinte+1:nbinw,nbinw+nbintt+nbinte+1:nbinw+nbinw) !offdiag
       fisher(nbinee+1:bin_no,1:nbinee)=covmat(nbinw+nbintt+nbinte+1:nbinw+nbinw,nbintt+nbinte+1:nbinw) !offdiag
  !All
  else if ((use_tt .eqv. .true.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbin
       allocate(fisher(nbin,nbin))
       fisher(:,:)=covmat(:,:)
  else
     write(*,*) 'Fail: no possible options chosen'
  endif

  !Invert covmat
  call dpotrf('U',bin_no,fisher,bin_no,info)
  if(info.ne.0)then
     print*, ' info in dpotrf =', info
     stop
  endif
  call dpotri('U',bin_no,fisher,bin_no,info)
  if(info.ne.0)then
     print*, ' info in dpotri =', info
     stop
  endif
  do i=1,bin_no
     do j=i,bin_no
        fisher(j,i)=fisher(i,j)
     enddo
  enddo

  !Read window functions
  call get_free_lun(lun)
  open(unit=lun,file=bbld_file,form='formatted',status='unknown',action='read')
  allocate(win_func_d(1:bmax_win,1:lmax_win)) !Defined over ACTPol's full ell range 
  do i=1,bmax_win
     read(lun,*) (win_func_d(i,il), il=2,lmax_win)
  enddo
  close(lun)

  call get_free_lun(lun)
  open(unit=lun,file=bblw_file,form='formatted',status='unknown',action='read')
  allocate(win_func_w(1:bmax_win,1:lmax_win)) !Defined over ACTPol's full ell range 
  do i=1,bmax_win
     read(lun,*) (win_func_w(i,il), il=2,lmax_win)
  enddo
  close(lun)

  end subroutine actpol_like_init
  
  ! ===========================================================================
  subroutine actpol_calc_like(aplike,cell_tt,cell_te,cell_ee,yp2)

  real(campc), dimension(2:) :: cell_tt,cell_ee,cell_te
  real(campc) :: cltt(2:lmax_win), clte(2:lmax_win), clee(2:lmax_win)
  real(campc) :: cl_tt_w(bmax),cl_te_w(bmax),cl_ee_w(bmax),cl_tt_d(bmax),cl_te_d(bmax),cl_ee_d(bmax)
  real(campc) :: aplike, yp2
  integer :: bin_no,lun,il,i,j,info
  real(campc), allocatable ::  Y(:), X_model(:)
  real(campc), allocatable :: ptemp(:)

  if (.not. allocated(Y)) then
     allocate(X_model(nbin),Y(nbin))
     X_model = 0
     Y = 0
  end if

  cltt(2:lmax_win)=0.d0 !Neglect above tt_lmax
  clte(2:lmax_win)=0.d0 
  clee(2:lmax_win)=0.d0 

  do i=2,tt_lmax
     cltt(i)=cell_tt(i)/real(i)/real(i+1.d0)*2.d0*PI
     clte(i)=cell_te(i)/real(i)/real(i+1.d0)*2.d0*PI
     clee(i)=cell_ee(i)/real(i)/real(i+1.d0)*2.d0*PI
  end do
  
  cl_tt_d(1:bmax)=MATMUL(win_func_d(1+2*bmax:3*bmax,2:lmax_win),cltt(2:lmax_win)) !use 150x150 windows
  cl_te_d(1:bmax)=MATMUL(win_func_d(1+6*bmax:7*bmax,2:lmax_win),clte(2:lmax_win))
  cl_ee_d(1:bmax)=MATMUL(win_func_d(1+9*bmax:10*bmax,2:lmax_win),clee(2:lmax_win))
  cl_tt_w(1:bmax)=MATMUL(win_func_w(1+2*bmax:3*bmax,2:lmax_win),cltt(2:lmax_win)) !use 150x150 windows
  cl_te_w(1:bmax)=MATMUL(win_func_w(1+6*bmax:7*bmax,2:lmax_win),clte(2:lmax_win))
  cl_ee_w(1:bmax)=MATMUL(win_func_w(1+9*bmax:10*bmax,2:lmax_win),clee(2:lmax_win))
 
  !Select ell range in theory 
  X_model(1:nbintt) = cl_tt_d(b0:b0-1+nbintt) !TT deep
  X_model(nbintt+1:nbintt+nbinte) = cl_te_d(1:nbinte)*yp2 !TE deep
  X_model(nbintt+nbinte+1:nbintt+nbinte+nbinee) = cl_ee_d(1:nbinee)*yp2*yp2 !EE deep
  X_model(nbintt+nbinte+nbinee+1:2*nbintt+nbinte+nbinee) = cl_tt_w(b0:b0-1+nbintt) !TT wide
  X_model(2*nbintt+nbinte+nbinee+1:2*nbintt+2*nbinte+nbinee) = cl_te_w(1:nbinte)*yp2 !TE wide
  X_model(2*nbintt+2*nbinte+nbinee+1:2*nbintt+2*nbinte+2*nbinee) = cl_ee_w(1:nbinee)*yp2*yp2 !EE wide
 
  !Start basic chisq 
  Y = X_data - X_model

  !Select data
  !Only TT
  if((use_tt .eqv. .true.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .false.)) then
       bin_no=nbintt+nbintt
       allocate(diff_vec(bin_no),ptemp(bin_no))
       diff_vec(1:nbintt)=Y(1:nbintt) !deep
       diff_vec(nbintt+1:bin_no)=Y(nbinw+1:nbinw+nbintt) !wide
  !Only TE
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .false.)) then 
       bin_no=nbinte+nbinte
       allocate(diff_vec(bin_no),ptemp(bin_no))
       diff_vec(1:nbinte)=Y(nbintt+1:nbintt+nbinte) !deep
       diff_vec(nbinte+1:bin_no)=Y(nbinw+nbintt+1:nbinw+nbintt+nbinte) !wide
  !Only EE
  else if((use_tt .eqv. .false.) .and. (use_te .eqv. .false.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbinee+nbinee
       allocate(diff_vec(bin_no),ptemp(bin_no))
       diff_vec(1:nbinee)=Y(nbintt+nbinte+1:nbintt+nbinte+nbinee) !deep
       diff_vec(nbinee+1:bin_no)=Y(nbinw+nbintt+nbinte+1:nbinw+nbintt+nbinte+nbinee) !wide
  !All
  else if ((use_tt .eqv. .true.) .and. (use_te .eqv. .true.) .and. (use_ee .eqv. .true.)) then
       bin_no=nbin
       allocate(diff_vec(nbin),ptemp(nbin))
       diff_vec(:)=Y(:)
  else
     write(*,*) 'Fail: no possible options chosen'
  endif

  ptemp=matmul(fisher,diff_vec)
  aplike=sum(ptemp*diff_vec)
  aplike = aplike/2.d0
 
  deallocate(X_model,Y)
  deallocate(diff_vec,ptemp)
    
 end subroutine actpol_calc_like
  
 subroutine get_free_lun(lun)

 implicit none
 integer, intent(out) :: lun
 integer, save :: last_lun = 19  
 logical :: used
 lun = last_lun
 do
   inquire( unit=lun, opened=used )
   if ( .not. used ) exit
       lun = lun + 1
 end do
    
 last_lun = lun
 end subroutine get_free_lun

end module ACTPol_CMBonly
